#include <bnb/glsl.frag>





BNB_IN(0) vec2 var_uv;
BNB_IN(1) vec2 var_bg_mask_uv;


BNB_DECLARE_SAMPLER_2D(2, 3, glfx_BACKGROUND);

BNB_DECLARE_SAMPLER_2D(0, 1, glfx_BG_MASK);

vec4 cubic(float v) {
    vec4 n = vec4(1.0, 2.0, 3.0, 4.0) - v;
    vec4 s = n * n * n;
    float x = s.x;
    float y = s.y - 4.0 * s.x;
    float z = s.z - 4.0 * s.y + 6.0 * s.x;
    float w = 6.0 - x - y - z;
    return vec4(x, y, z, w) * (1.0 / 6.0);
}

float get_bg_mask(){
    vec2 bg_uv = var_bg_mask_uv;
    const float threshold = 0.2;

    vec2 texSize = background_nn_meta.xy;
    
    vec2 invTexSize = 1.0 / texSize;

    vec2 texCoords = bg_uv * texSize - 0.5;

    vec2 fxy = fract(texCoords);
    texCoords -= fxy;

    vec4 xcubic = cubic(fxy.x);
    vec4 ycubic = cubic(fxy.y);

    vec4 c = texCoords.xxyy + vec2(-0.5, +1.5).xyxy;

    vec4 s = vec4(xcubic.xz + xcubic.yw, ycubic.xz + ycubic.yw);
    vec4 offset = c + vec4(xcubic.yw, ycubic.yw) / s;

    offset *= invTexSize.xxyy;

    vec4 sample0 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.xz);
    vec4 sample1 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.yz);
    vec4 sample2 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.xw);
    vec4 sample3 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.yw);

    float sx = s.x / (s.x + s.y);
    float sy = s.z / (s.z + s.w);

    vec4 filtered_mask = mix(mix(sample3, sample2, sx), mix(sample1, sample0, sx), sy);
    
    return max((filtered_mask.x - threshold) / (1.0 - threshold), 0.0);
}

vec2 rgb_hs(vec3 rgb)
{
    float cmax = max(rgb.r, max(rgb.g, rgb.b));
    float cmin = min(rgb.r, min(rgb.g, rgb.b));
    float delta = cmax - cmin;
    vec2 hs = vec2(0.0);

    if (cmax > cmin) {
        hs.y = delta/cmax;
        if (rgb.r == cmax)
            hs.x = (rgb.g - rgb.b) / delta;
        else 
        {
            if (rgb.g == cmax)
                hs.x = 2.0 + (rgb.b - rgb.r) / delta;
            else
                hs.x = 4.0 + (rgb.r - rgb.g) / delta;
        }
        hs.x = fract(hs.x / 6.0);
    }
    
    return hs;
}

float rgb_v(vec3 rgb)
{
    return max(rgb.r, max(rgb.g, rgb.b));
}

vec3 hsv_rgb(float h, float s, float v)
{
    return v * mix(vec3(1.0), clamp(abs(fract(vec3(1.0, 2.0 / 3.0, 1.0 / 3.0) + h) * 6.0 - 3.0) - 1.0, 0.0, 1.0), s);
}

float filtered_bg_simple( BNB_DECLARE_SAMPLER_2D_ARGUMENT(mask_tex), vec2 uv )
{
	float bg1 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(mask_tex), uv ).x;
	if( bg1 > 0.98 || bg1 < 0.02 )
		return bg1;

	vec2 o = 1./vec2(textureSize(BNB_SAMPLER_2D(mask_tex),0));
	float bg2 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(mask_tex), uv + vec2(o.x,0.) ).x;
	float bg3 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(mask_tex), uv - vec2(o.x,0.) ).x;
	float bg4 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(mask_tex), uv + vec2(0.,o.y) ).x;
	float bg5 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(mask_tex), uv - vec2(0.,o.y) ).x;

	return 0.2*(bg1+bg2+bg3+bg4+bg5);
}

vec2 rotate_uv(vec2 uv, float angle)
{
	float aspect_ratio = bnb_SCREEN.y / bnb_SCREEN.x;
	float c = cos(radians(angle));
	float s = sin(radians(angle));

	uv.y *= aspect_ratio;

	uv = vec2(mat3(c, -s, 0., s, c, 0., 0.5, 0.5, 1.0) * vec3(uv - 0.5, 1.));

	uv.y *= 1. / aspect_ratio;
	return uv;
}

vec2 scale_uv(vec2 uv, float scale)
{
	float inv_scale = 1. / scale;
	return vec2(mat3(inv_scale, 0., 0., 0., inv_scale, 0., 0.5, 0.5, 1.0) * vec3(uv - 0.5, 1.));
}

vec3 blendColor(vec3 base, vec3 blend) {
    float v = rgb_v(base);
    vec2 hs = rgb_hs(blend);
    return hsv_rgb(hs.x, hs.y, v);
}

vec3 blendColor(vec3 base, vec3 blend, float opacity) {
    return (blendColor(base, blend) * opacity + base * (1.0 - opacity));
}

void main()
{   
    vec2 uv = var_uv;

    #ifdef BNB_VK_1
        uv.y = 1. - uv.y;
    #endif
    
    vec4 bg = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BACKGROUND), uv);

	uv.y = 1. - uv.y;

    float mask = get_bg_mask();

    bnb_FragColor.rgb = bg.rgb;

    bnb_FragColor.a = 1.-mask*js_bg_alpha.x;
    
    bnb_FragColor.rgb *= bnb_FragColor.a;
}
