#include <bnb/glsl.frag>


BNB_IN(0) vec4 var_uv;



BNB_DECLARE_SAMPLER_2D(2, 3, tex_df_blur);

BNB_DECLARE_SAMPLER_2D(4, 5, glfx_BACKGROUND);

BNB_DECLARE_SAMPLER_2D(0, 1, glfx_BG_MASK);

vec4 cubic(float v) {
    vec4 n = vec4(1.0, 2.0, 3.0, 4.0) - v;
    vec4 s = n * n * n;
    float x = s.x;
    float y = s.y - 4.0 * s.x;
    float z = s.z - 4.0 * s.y + 6.0 * s.x;
    float w = 6.0 - x - y - z;
    return vec4(x, y, z, w) * (1.0 / 6.0);
}

float get_bg_mask(){
    vec2 bg_uv = var_uv.zw;

    const float threshold = 0.2;

    vec2 texSize = background_nn_meta.xy;
    
    vec2 invTexSize = 1.0 / texSize;

    vec2 texCoords = bg_uv * texSize - 0.5;

    vec2 fxy = fract(texCoords);
    texCoords -= fxy;

    vec4 xcubic = cubic(fxy.x);
    vec4 ycubic = cubic(fxy.y);

    vec4 c = texCoords.xxyy + vec2(-0.5, +1.5).xyxy;

    vec4 s = vec4(xcubic.xz + xcubic.yw, ycubic.xz + ycubic.yw);
    vec4 offset = c + vec4(xcubic.yw, ycubic.yw) / s;

    offset *= invTexSize.xxyy;

    vec4 sample0 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.xz);
    vec4 sample1 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.yz);
    vec4 sample2 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.xw);
    vec4 sample3 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.yw);

    float sx = s.x / (s.x + s.y);
    float sy = s.z / (s.z + s.w);

    vec4 filtered_mask = mix(mix(sample3, sample2, sx), mix(sample1, sample0, sx), sy);
    
    return max((filtered_mask.x - threshold) / (1.0 - threshold), 0.0);
}

void main()
{   
    vec2 uv = var_uv.xy;
	vec3 blurred = BNB_TEXTURE_2D(BNB_SAMPLER_2D(tex_df_blur),uv).xyz;
    #ifdef BNB_VK_1
        uv.y = 1. - uv.y;
    #endif
    vec3 bg = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BACKGROUND),uv).xyz;

    float mask = get_bg_mask();

	bnb_FragColor = vec4( mix(bg,blurred, mask),1. );
}
