#include <bnb/glsl.frag>





BNB_IN(0) vec2 var_uv;
BNB_IN(1) vec2 var_bg_mask_uv;


BNB_DECLARE_SAMPLER_2D(0, 1, bgTex);

BNB_DECLARE_SAMPLER_VIDEO(6, 7, glfx_VIDEO);

BNB_DECLARE_SAMPLER_2D(4, 5, glfx_BACKGROUND);

BNB_DECLARE_SAMPLER_2D(2, 3, glfx_BG_MASK);

vec4 cubic(float v) {
    vec4 n = vec4(1.0, 2.0, 3.0, 4.0) - v;
    vec4 s = n * n * n;
    float x = s.x;
    float y = s.y - 4.0 * s.x;
    float z = s.z - 4.0 * s.y + 6.0 * s.x;
    float w = 6.0 - x - y - z;
    return vec4(x, y, z, w) * (1.0 / 6.0);
}

vec2 rgb_hs(vec3 rgb)
{
    float cmax = max(rgb.r, max(rgb.g, rgb.b));
    float cmin = min(rgb.r, min(rgb.g, rgb.b));
    float delta = cmax - cmin;
    vec2 hs = vec2(0.0);

    if (cmax > cmin) {
        hs.y = delta/cmax;
        if (rgb.r == cmax)
            hs.x = (rgb.g - rgb.b) / delta;
        else 
        {
            if (rgb.g == cmax)
                hs.x = 2.0 + (rgb.b - rgb.r) / delta;
            else
                hs.x = 4.0 + (rgb.r - rgb.g) / delta;
        }
        hs.x = fract(hs.x / 6.0);
    }
    
    return hs;
}

float rgb_v(vec3 rgb)
{
    return max(rgb.r, max(rgb.g, rgb.b));
}

vec3 hsv_rgb(float h, float s, float v)
{
    return v * mix(vec3(1.0), clamp(abs(fract(vec3(1.0, 2.0 / 3.0, 1.0 / 3.0) + h) * 6.0 - 3.0) - 1.0, 0.0, 1.0), s);
}

vec2 rotate_uv(vec2 uv, float angle)
{
    if (angle == 0.) return uv;

    float c = cos(radians(angle));
    float s = sin(radians(angle));
    
    uv = vec2(mat3(c, -s, 0., s, c, 0., 0.5, 0.5, 1.0) * vec3(uv - 0.5, 1.));
    
    return uv;
}

vec2 scale_uv_tex(vec2 uv,BNB_DECLARE_SAMPLER_2D_ARGUMENT(bgTex))
{
    float angle = js_bg_rotation.x;

    #ifndef BNB_GL_ES_1
        vec2 textureSize = vec2(textureSize(BNB_SAMPLER_2D(bgTex), 0));
    #else
        vec2 textureSize = js_texture_size.xy;
    #endif

    float aspect_ratio = bnb_SCREEN.y / bnb_SCREEN.x;
    float texture_aspect_ratio = textureSize.y / textureSize.x;
    float scale_x = 1.0;
    float scale_y = 1.0;

    if (angle == 90. || angle == -90. || angle == 270. || angle == -270.){
        if (texture_aspect_ratio < 1. && texture_aspect_ratio > aspect_ratio) {
            scale_x = 1. / texture_aspect_ratio / aspect_ratio;
        } else {
            scale_x =  1./ aspect_ratio / texture_aspect_ratio;
            if (scale_x < 1.) {
                scale_y /= scale_x;
                scale_x /= scale_x;
            }
        } 
    }
    else {
        if (texture_aspect_ratio > aspect_ratio) {
            scale_y = texture_aspect_ratio / aspect_ratio;
        } else {
            scale_x = aspect_ratio / texture_aspect_ratio;
        }
    }

    float inv_scale_x = 1. / scale_x;
    float inv_scale_y = 1. / scale_y;
    
    return vec2(mat3(inv_scale_x, 0., 0., 0., inv_scale_y, 0., 0.5, 0.5, 1.0) * vec3(uv - 0.5, 1.));
}

vec3 blendColor(vec3 base, vec3 blend) {
    float v = rgb_v(base);
    vec2 hs = rgb_hs(blend);
    return hsv_rgb(hs.x, hs.y, v);
}

vec3 blendColor(vec3 base, vec3 blend, float opacity) {
    return (blendColor(base, blend) * opacity + base * (1.0 - opacity));
}

float get_bg_mask(){
    vec2 bg_uv = var_bg_mask_uv;
    
    const float threshold = 0.2;

    vec2 texSize = background_nn_meta.xy;
    
    vec2 invTexSize = 1.0 / texSize;

    vec2 texCoords = bg_uv * texSize - 0.5;

    vec2 fxy = fract(texCoords);
    texCoords -= fxy;

    vec4 xcubic = cubic(fxy.x);
    vec4 ycubic = cubic(fxy.y);

    vec4 c = texCoords.xxyy + vec2(-0.5, +1.5).xyxy;

    vec4 s = vec4(xcubic.xz + xcubic.yw, ycubic.xz + ycubic.yw);
    vec4 offset = c + vec4(xcubic.yw, ycubic.yw) / s;

    offset *= invTexSize.xxyy;

    vec4 sample0 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.xz);
    vec4 sample1 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.yz);
    vec4 sample2 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.xw);
    vec4 sample3 = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BG_MASK), offset.yw);

    float sx = s.x / (s.x + s.y);
    float sy = s.z / (s.z + s.w);

    vec4 filtered_mask = mix(mix(sample3, sample2, sx), mix(sample1, sample0, sx), sy);
    
    return max((filtered_mask.x - threshold) / (1.0 - threshold), 0.0);
}

void main()
{
    vec2 uv = var_uv;
    
    float screenRatio_y = bnb_SCREEN.y/bnb_SCREEN.x;
    float screenRatio_x = bnb_SCREEN.x/bnb_SCREEN.y;

    vec4 bg = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_BACKGROUND),uv);
    
    float angle = js_bg_rotation.x;
    
    uv = rotate_uv(uv,angle);
    if (bg_plane_mode.x == 0.){
        uv = scale_uv_tex(uv,BNB_PASS_SAMPLER_ARGUMENT(bgTex));
    } else {
        uv = scale_uv_tex(uv,BNB_PASS_SAMPLER_ARGUMENT(glfx_VIDEO));
    }

    #if defined(BNB_GL_ES_3) || defined(BNB_GL)
        uv.y = 1. - uv.y;
    #endif

    #ifdef BNB_GL_ES_1
         uv.y = 1. - uv.y;
    #endif
    
    vec3 bg_color;
    if (bg_plane_mode.x == 0.){
        bg_color = BNB_TEXTURE_2D(BNB_SAMPLER_2D(bgTex), uv).xyz;
    } else {
        bg_color = BNB_TEXTURE_2D(BNB_SAMPLER_2D(glfx_VIDEO), uv).xyz;
    }
    
    float mask = get_bg_mask();
    
    bnb_FragColor = vec4(mix(bg.rgb, bg_color, mask * bgTextureAlpha.x),1.);
}